"""Kubernetes adaptors"""

# pylint: disable=import-outside-toplevel

import functools
import os

from sky.utils import env_options
from sky.utils import ux_utils

kubernetes = None
urllib3 = None

_configured = False
_core_api = None
_auth_api = None
_networking_api = None
_custom_objects_api = None
_node_api = None

# Timeout to use for API calls
API_TIMEOUT = 5


def import_package(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global kubernetes
        global urllib3
        if kubernetes is None:
            try:
                import kubernetes as _kubernetes
                import urllib3 as _urllib3
            except ImportError:
                # TODO(romilb): Update this message to point to installation
                #  docs when they are ready.
                raise ImportError('Fail to import dependencies for Kubernetes. '
                                  'Run `pip install kubernetes` to '
                                  'install them.') from None
            kubernetes = _kubernetes
            urllib3 = _urllib3
        return func(*args, **kwargs)

    return wrapper


@import_package
def get_kubernetes():
    return kubernetes


@import_package
def _load_config():
    global _configured
    if _configured:
        return
    try:
        # Load in-cluster config if running in a pod
        # Kubernetes set environment variables for service discovery do not
        # show up in SkyPilot tasks. For now, we work around by using
        # DNS name instead of environment variables.
        # See issue: https://github.com/skypilot-org/skypilot/issues/2287
        os.environ['KUBERNETES_SERVICE_HOST'] = 'kubernetes.default.svc'
        os.environ['KUBERNETES_SERVICE_PORT'] = '443'
        kubernetes.config.load_incluster_config()
    except kubernetes.config.config_exception.ConfigException:
        try:
            kubernetes.config.load_kube_config()
        except kubernetes.config.config_exception.ConfigException as e:
            suffix = ''
            if env_options.Options.SHOW_DEBUG_INFO.get():
                suffix += f' Error: {str(e)}'
            # Check if exception was due to no current-context
            if 'Expected key current-context' in str(e):
                err_str = ('Failed to load Kubernetes configuration. '
                           'Kubeconfig does not contain any valid context(s).'
                           f'{suffix}\n'
                           '    If you were running a local Kubernetes '
                           'cluster, run `sky local up` to start the cluster.')
            else:
                err_str = (
                    'Failed to load Kubernetes configuration. '
                    f'Please check if your kubeconfig file is valid.{suffix}')
            err_str += '\nTo disable Kubernetes for SkyPilot: run `sky check`.'
            with ux_utils.print_exception_no_traceback():
                raise ValueError(err_str) from None
    _configured = True


@import_package
def core_api():
    global _core_api
    if _core_api is None:
        _load_config()
        _core_api = kubernetes.client.CoreV1Api()

    return _core_api


@import_package
def auth_api():
    global _auth_api
    if _auth_api is None:
        _load_config()
        _auth_api = kubernetes.client.RbacAuthorizationV1Api()

    return _auth_api


@import_package
def networking_api():
    global _networking_api
    if _networking_api is None:
        _load_config()
        _networking_api = kubernetes.client.NetworkingV1Api()

    return _networking_api


@import_package
def custom_objects_api():
    global _custom_objects_api
    if _custom_objects_api is None:
        _load_config()
        _custom_objects_api = kubernetes.client.CustomObjectsApi()

    return _custom_objects_api


@import_package
def node_api():
    global _node_api
    if _node_api is None:
        _load_config()
        _node_api = kubernetes.client.NodeV1Api()

    return _node_api


@import_package
def api_exception():
    return kubernetes.client.rest.ApiException


@import_package
def config_exception():
    return kubernetes.config.config_exception.ConfigException


@import_package
def max_retry_error():
    return urllib3.exceptions.MaxRetryError


@import_package
def stream():
    return kubernetes.stream.stream



"""GCP cloud adaptors"""

# pylint: disable=import-outside-toplevel
import functools
import json

googleapiclient = None
google = None


def import_package(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global googleapiclient, google
        if googleapiclient is None or google is None:
            try:
                import google as _google
                import googleapiclient as _googleapiclient
                googleapiclient = _googleapiclient
                google = _google
            except ImportError:
                raise ImportError('Failed to import dependencies for GCP. '
                                  'Try: pip install "skypilot[gcp]"') from None
        return func(*args, **kwargs)

    return wrapper


@import_package
def build(service_name: str, version: str, *args, **kwargs):
    """Build a GCP service.

    Args:
        service_name: GCP service name (e.g., 'compute', 'storagetransfer').
        version: Service version (e.g., 'v1').
    """
    from googleapiclient import discovery
    return discovery.build(service_name, version, *args, **kwargs)


@import_package
def storage_client():
    """Helper method that connects to GCS Storage Client for
    GCS Bucket
    """
    from google.cloud import storage
    return storage.Client()


@import_package
def anonymous_storage_client():
    """Helper method that connects to GCS Storage Client for
    Public GCS Buckets
    """
    from google.cloud import storage
    return storage.Client.create_anonymous_client()


@import_package
def not_found_exception():
    """NotFound exception."""
    from google.api_core import exceptions as gcs_exceptions
    return gcs_exceptions.NotFound


@import_package
def forbidden_exception():
    """Forbidden exception."""
    from google.api_core import exceptions as gcs_exceptions
    return gcs_exceptions.Forbidden


@import_package
def http_error_exception():
    """HttpError exception."""
    from googleapiclient import errors
    return errors.HttpError


@import_package
def credential_error_exception():
    """CredentialError exception."""
    from google.auth import exceptions
    return exceptions.DefaultCredentialsError


@import_package
def get_credentials(cred_type: str, credentials_field: str):
    """Get GCP credentials."""
    from google.oauth2 import service_account
    from google.oauth2.credentials import Credentials as OAuthCredentials

    if cred_type == 'service_account':
        # If parsing the gcp_credentials failed, then the user likely made a
        # mistake in copying the credentials into the config yaml.
        try:
            service_account_info = json.loads(credentials_field)
        except json.decoder.JSONDecodeError as e:
            raise RuntimeError('gcp_credentials found in cluster yaml file but '
                               'formatted improperly.') from e
        credentials = service_account.Credentials.from_service_account_info(
            service_account_info)
    elif cred_type == 'credentials_token':
        # Otherwise the credentials type must be credentials_token.
        credentials = OAuthCredentials(credentials_field)
    return credentials



""" data fetcher for vsphere """
import json
import logging
import os
import typing

import pandas as pd

from sky.adaptors import vsphere as vsphere_adaptor
from sky.clouds.service_catalog.common import get_catalog_path
from sky.provision.vsphere.common.cls_api_client import ClsApiClient

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# TODO: Add more GPUs and a GPU full/short name mapping
PRESET_GPU_LIST = [
    {
        'Model': 'V100',
        'Type': 'GPU',
        'MemoryMB': 1024 * 16,
        'vCPUs': 8,
        'fullNames': [
            'Tesla V100',
            'Tesla V100-PCIE-16GB',
            'Tesla V100-PCIE-32GB',
            'Tesla V100-SXM2-16GB',
        ],
    },
    {
        'Model': 'P100',
        'Type': 'GPU',
        'MemoryMB': 1024 * 16,
        # Estimated
        'vCPUs': 8,
        'fullNames': [
            'Tesla P100-PCIE-16GB',
            'Tesla P100-PCIE-12GB',
            'Tesla P100-SXM2-16GB',
            'Tesla P100-SXM2-12GB',
        ],
    },
    {
        'Model': 'T4',
        'Type': 'GPU',
        'MemoryMB': 1024 * 16,
        'vCPUs': 4,
        'fullNames': ['Tesla T4'],
    },
    {
        'Model': 'P4',
        'Type': 'GPU',
        'MemoryMB': 1024 * 8,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['Tesla P4'],
    },
    {
        'Model': 'K80',
        'Type': 'GPU',
        'MemoryMB': 1024 * 24,
        'vCPUs': 8,
        'fullNames': ['Tesla K80'],
    },
    {
        'Model': 'A100',
        'Type': 'GPU',
        'MemoryMB': 1024 * 40,
        'vCPUs': 8,
        'fullNames': ['Tesla A100-PCIE-40GB', 'Tesla A100-SXM4-40GB'],
    },
    {
        'Model': '1080',
        'Type': 'GPU',
        'MemoryMB': 1024 * 8,
        # Estimated
        'vCPUs': 4,
        'fullNames': ['GeForce GTX 1080 Ti', 'GeForce GTX 1080'],
    },
    {
        'Model': '2080',
        'Type': 'GPU',
        'MemoryMB': 1024 * 8,
        # Estimated
        'vCPUs': 4,
        'fullNames': [
            'GeForce RTX 2080 Ti',
            'GeForce RTX 2080 SUPER',
            'GeForce RTX 2080',
        ],
    },
    {
        'Model': 'A5000',
        'Type': 'GPU',
        'MemoryMB': 1024 * 24,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['NVIDIA RTX A5000'],
    },
    {
        'Model': 'A6000',
        'Type': 'GPU',
        'MemoryMB': 1024 * 48,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['NVIDIA RTX A6000'],
    },
    {
        'Model': 'K1200',
        'Type': 'GPU',
        'MemoryMB': 1024 * 4,
        # Estimated
        'vCPUs': 2,
        'fullNames': ['Quadro K1200'],
    },
    {
        'Model': 'M40',
        'Type': 'GPU',
        'MemoryMB': 1024 * 24,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['Quadro M40'],
    },
    {
        'Model': 'M60',
        'Type': 'GPU',
        'MemoryMB': 1024 * 8,
        # Estimated
        'vCPUs': 4,
        'fullNames': ['Quadro M60'],
    },
    {
        'Model': 'M5000',
        'Type': 'GPU',
        'MemoryMB': 1024 * 8,
        # Estimated
        'vCPUs': 4,
        'fullNames': ['Quadro M5000'],
    },
    {
        'Model': 'M4000',
        'Type': 'GPU',
        'MemoryMB': 1024 * 8,
        # Estimated
        'vCPUs': 4,
        'fullNames': ['Quadro M4000'],
    },
    {
        'Model': 'Titan X',
        'Type': 'GPU',
        'MemoryMB': 1024 * 12,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['GeForce GTX TITAN X', 'GeForce GTX TITAN X (Pascal)'],
    },
    {
        'Model': 'Titan V',
        'Type': 'GPU',
        'MemoryMB': 1024 * 12,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['TITAN V'],
    },
    {
        'Model': 'P5000',
        'Type': 'GPU',
        'MemoryMB': 1024 * 16,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['Quadro P5000'],
    },
    {
        'Model': 'P6000',
        'Type': 'GPU',
        'MemoryMB': 1024 * 24,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['Quadro P6000'],
    },
    {
        'Model': 'RTX 6000',
        'Type': 'GPU',
        'MemoryMB': 1024 * 24,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['NVIDIA RTX 6000'],
    },
    {
        'Model': '2080 Ti',
        'Type': 'GPU',
        'MemoryMB': 1024 * 11,
        # Estimated
        'vCPUs': 4,
        'fullNames': ['GeForce RTX 2080 Ti'],
    },
    {
        'Model': '3070',
        'Type': 'GPU',
        'MemoryMB': 1024 * 8,
        # Estimated
        'vCPUs': 4,
        'fullNames': ['GeForce RTX 3070'],
    },
    {
        'Model': '3080',
        'Type': 'GPU',
        'MemoryMB': 1024 * 10,
        # Estimated
        'vCPUs': 4,
        'fullNames': ['GeForce RTX 3080'],
    },
    {
        'Model': '3090',
        'Type': 'GPU',
        'MemoryMB': 1024 * 24,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['GeForce RTX 3090'],
    },
    {
        'Model': '4090',
        'Type': 'GPU',
        'MemoryMB': 1024 * 24,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['GeForce RTX 4090'],
    },
    {
        'Model': '4080',
        'Type': 'GPU',
        'MemoryMB': 1024 * 16,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['GeForce RTX 4080'],
    },
    {
        'Model': '4070',
        'Type': 'GPU',
        'MemoryMB': 1024 * 12,
        # Estimated
        'vCPUs': 8,
        'fullNames': ['GeForce RTX 4070'],
    },
    {
        'Model': 'G200eW3',
        'Type': 'GPU',
        'MemoryMB': 1024 * 12,
        'vCPUs': 8,
        'fullNames': ['Integrated Matrox G200eW3 Graphics Controller'],
    },
    {
        'Model': 'K1200',
        'Type': 'GPU',
        'MemoryMB': 1024 * 4,
        'vCPUs': 2,
        'fullNames': ['GM107GL [Quadro K1200]'],
    },
]
ACCELERATORS_CSV_PATH = get_catalog_path('vsphere/accelerators.csv')
CPU_INSTANCE_TYPES = [
    {
        'InstanceType': 'cpu.xsmall',
        'vCPUs': 2,
        'MemoryMB': 1024 * 2,
    },
    {
        'InstanceType': 'cpu.small',
        'vCPUs': 2,
        'MemoryMB': 1024 * 4,
    },
    {
        'InstanceType': 'cpu.medium',
        'vCPUs': 2,
        'MemoryMB': 1024 * 8,
    },
    {
        'InstanceType': 'cpu.large',
        'vCPUs': 4,
        'MemoryMB': 1024 * 16,
    },
    {
        'InstanceType': 'cpu.xlarge',
        'vCPUs': 4,
        'MemoryMB': 1024 * 32,
    },
    {
        'InstanceType': 'cpu.2xlarge',
        'vCPUs': 8,
        'MemoryMB': 1024 * 64,
    },
    {
        'InstanceType': 'cpu.4xlarge',
        'vCPUs': 16,
        'MemoryMB': 1024 * 128,
    },
    {
        'InstanceType': 'cpu.8xlarge',
        'vCPUs': 32,
        'MemoryMB': 1024 * 128,
    },
]
GPU_MANUFACTURERS = ['nvidia', 'amd', 'intel']


def initialize_accelerators_csv():
    if not os.path.exists(ACCELERATORS_CSV_PATH):
        try:
            # Create an empty DataFrame with the desired columns
            data = [{
                'Model': gpu.get('Model'),
                'Type': gpu.get('Type'),
                'MemoryMB': gpu.get('MemoryMB'),
                'vCPUs': gpu.get('vCPUs'),
                'fullNames': gpu.get('fullNames'),
            } for gpu in PRESET_GPU_LIST]
            df = pd.DataFrame(data)

            # Write DataFrame to CSV
            df.to_csv(ACCELERATORS_CSV_PATH, index=False)
            return
        except Exception as e:
            logger.error(f'Failed to initialize accelerators.csv: {e}')
            raise e


def get_accelerators_from_csv() -> typing.List[typing.Dict]:
    try:
        df = pd.read_csv(ACCELERATORS_CSV_PATH)
        return df.to_dict('records')
    except Exception as e:
        logger.error(f'Failed to get accelerators from accelerators.csv: {e}')
        raise e


def initialize_hosts_csv(csv_saving_path: str,
                         hosts_for_csv: typing.List[typing.Dict]):
    try:
        with open(csv_saving_path, 'a', encoding='utf-8') as f:
            for host in hosts_for_csv:
                f.write(f'{host.get("HostName")},{host.get("MoRefID")},'
                        f'{host.get("vCenter")},{host.get("Datacenter")},'
                        f'{host.get("Cluster")},{host.get("TotalCPUMHz")},'
                        f'{host.get("AvailableCPUCapacityMHz")},'
                        f'{host.get("TotalMemoryMB")},'
                        f'{host.get("AvailableMemoryMB")},'
                        f'"{host.get("Accelerators")}",'
                        f'{host.get("CPUMHz")},{host.get("UUID")}\n')
                return
    except Exception as e:
        logger.error(f'Failed to initialize hosts.csv: {e}')
        raise e


def summarize_by_cluster_and_gpu(
    hosts_to_summarize: typing.List[typing.Dict],
) -> typing.Dict[str, typing.Any]:
    try:
        gpu_cluster_dict: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
        gpu_list = get_accelerators_from_csv()
        weakest_host_of_each_cluster = {}
        for host in hosts_to_summarize:
            # If cluster is not in gpu_cluster_dict, initialize cluster entry
            cluster_name = host.get('Cluster', '')

            #  Deal with host with gpus
            if host.get('Accelerators') != []:
                for accelerator in host.get('Accelerators', []):
                    gpu_key = accelerator.get('DeviceName', '')
                    # if accelerator.get('Status') == 'Available':
                    # we just list all the instances we supported there
                    # no matter whether
                    # the GPU is available or not.
                    if cluster_name not in gpu_cluster_dict:
                        gpu_cluster_dict[cluster_name] = {}
                    if gpu_key not in gpu_cluster_dict[cluster_name]:
                        # Go through PRESET_GPU_LIST to find the GPU's
                        # memory and vCPU
                        # TODO: Confirm if 1:1 mapping between VM memory and
                        #  GPU memory is good.
                        for gpu in gpu_list:
                            model = gpu.get('Model')
                            if model and isinstance(model, str):
                                if model.lower() in gpu_key.lower():
                                    gpu_cluster_dict[cluster_name][gpu_key] = {
                                        'AcceleratorCount': 1,
                                        'vCPUs': gpu.get('vCPUs', 0),
                                        'MemoryMB': gpu.get('MemoryMB', 0),
                                        'Accelerators': accelerator,
                                        'GpuInfo': host.get('GpuInfo', {}),
                                    }
                                    break
            # Try to compare its capacity with the weakest CPU host of
            # the cluster
            if cluster_name not in weakest_host_of_each_cluster:
                weakest_host_of_each_cluster[cluster_name] = host.get(
                    'TotalCPUcores', 0)
            else:
                if (host.get('TotalCPUcores') and host.get('TotalCPUcores') <
                        weakest_host_of_each_cluster[cluster_name]):
                    weakest_host_of_each_cluster[cluster_name] = host.get(
                        'TotalCPUcores', 0)
        return {
            'gpu_cluster_dict': gpu_cluster_dict,
            'weakest_host_of_each_cluster': weakest_host_of_each_cluster,
        }
    except Exception as e:
        logger.error(f'Failed to summarize hosts: {e}')
        raise e


def initialize_vms_csv(csv_saving_path: str,
                       hosts_for_csv: typing.List[typing.Dict],
                       vcenter_name: str) -> None:
    try:
        with open(csv_saving_path, 'a', encoding='utf-8') as f:
            # Read info from summarize_by_cluster_and_gpu(hosts)
            summary_result = summarize_by_cluster_and_gpu(hosts_for_csv)
            gpu_clusters_dict = summary_result.get('gpu_cluster_dict')
            if gpu_clusters_dict:
                for cluster_name in gpu_clusters_dict:
                    gpus_dict = gpu_clusters_dict[cluster_name]
                    if gpus_dict:
                        for gpu_name in gpus_dict:
                            gpu_short_name = gpu_name
                            for gpu in PRESET_GPU_LIST:
                                model = gpu.get('Model')
                                if model and isinstance(model, str):
                                    if model.lower() in gpu_name.lower():
                                        gpu_short_name = gpu.get('Model')
                            if gpu_short_name == gpu_name:
                                logger.warning(
                                    f'Failed to find GPU {gpu_name}\'s '
                                    f'short name in the preset GPU list. '
                                    f'Use its full name instead.')
                            gpu_info = gpus_dict[gpu_name]
                            # TODO: add support for instance name with
                            #  gpu count, like gpu.A100.4
                            accelerator_count = gpu_info.get(
                                'AcceleratorCount', 0)
                            vcpus = gpu_info.get('vCPUs', 0)
                            memory_gb = round(
                                gpu_info.get('MemoryMB') / 1024, 2)
                            gpu_info_str = gpu_info['GpuInfo']
                            f.write(f'gpu.{gpu_short_name},{gpu_short_name},'
                                    f'{accelerator_count},'
                                    f'{vcpus},'
                                    f'{memory_gb},'
                                    f'\"{gpu_info_str}\",0.0,0.0,'
                                    f'{vcenter_name},{cluster_name}\n')
            cpu_host_dict = summary_result.get('weakest_host_of_each_cluster')
            if cpu_host_dict:
                for cluster_name in cpu_host_dict:
                    minimal_cpu_cores = cpu_host_dict[cluster_name]
                    # find the instance type that match the host's capacity
                    for instance_type in CPU_INSTANCE_TYPES:
                        if instance_type.get('vCPUs') <= minimal_cpu_cores:
                            vcpus = instance_type.get('vCPUs')
                            memory_mb = (instance_type.get('MemoryMB', 0))
                            # Make sure that memoryMB is a number before
                            # proceeding
                            if isinstance(memory_mb, (int, float)):
                                memory_in_gb = round(memory_mb / 1024, 2)
                            else:
                                raise TypeError(
                                    f'The value for \'MemoryMB\' is not a '
                                    f'number: {memory_mb}')
                            instance_type_str = instance_type.get(
                                'InstanceType')
                            f.write(f'{instance_type_str},,,'
                                    f'{vcpus},{memory_in_gb},,0.0,0.0,'
                                    f'{vcenter_name},{cluster_name}\n')
        return
    except Exception as e:
        logger.error(f'Failed to initialize vms.csv: {e}')
        raise e


def initialize_images_csv(csv_saving_path: str, vc_object,
                          vcenter_name: str) -> None:
    vc_servicemanager = vc_object.servicemanager
    vc_stub_config = vc_servicemanager.stub_config
    vc_tag_obj = vsphere_adaptor.get_tagging_client().Tag(vc_stub_config)
    vc_tag_association_obj = vsphere_adaptor.get_tagging_client(
    ).TagAssociation(vc_stub_config)
    cl_client_obj = ClsApiClient(vc_servicemanager)
    content_library_ids = cl_client_obj.local_library_service.list()
    library_item_service_item_obj = cl_client_obj.library_item_service
    vmtx_service_item_obj = cl_client_obj.vmtx_service
    with open(csv_saving_path, 'a', encoding='utf-8') as f:
        for lib_id in content_library_ids:
            library_obj = cl_client_obj.local_library_service.get(
                library_id=lib_id)
            library_item_ids = library_item_service_item_obj.list(
                library_id=library_obj.id)
            for item_id in library_item_ids:
                item = library_item_service_item_obj.get(
                    library_item_id=item_id)
                if item.type == 'vm-template':
                    item_spec = vmtx_service_item_obj.get(
                        template_library_item=item_id)
                    item_cpu = item_spec.cpu.count
                    item_memory = item_spec.memory.size_mib
                elif item.type in ['ovf', 'ova']:
                    item_cpu = 2
                    item_memory = 2048
                else:
                    continue
                dynamic_id = vsphere_adaptor.get_std_client().DynamicID(
                    type='com.vmware.content.library.Item', id=item.id)
                # Retrieve the tag_ids associated with this template item
                associated_tags = vc_tag_association_obj.list_attached_tags(
                    object_id=dynamic_id)
                # For each tag_id, get the tag details
                gpu_tags = []
                # Noted that if an image is tagged with more than one CPU tags,
                # it will be listed multiple times in
                # images.csv, so I added this logic.
                cpu_image_added = False
                for tag_id in associated_tags:
                    tag = vc_tag_obj.get(tag_id)
                    tag_name = tag.name
                    if 'SKYPILOT-CPU' in tag_name.upper():
                        if not cpu_image_added:
                            f.write(f'{item.id},{vcenter_name},{item_cpu},'
                                    f'{item_memory},,,\'[]\'\n')
                            cpu_image_added = True
                    elif 'GPU-' in tag_name.upper() or (
                            'SKYPILOT-' in tag_name.upper() and
                            'SKYPILOT-CPU' not in tag_name.upper()):
                        gpu_name = tag_name.split('-')[1]
                        if gpu_name not in gpu_tags:
                            gpu_tags.append(gpu_name)
                if len(gpu_tags) > 0:
                    gpu_tags_str = str(gpu_tags).replace('\'', '\"')
                    f.write(f'{item.id},{vcenter_name},{item_cpu},{item_memory}'
                            f',,,\'{gpu_tags_str}\'\n')


def initialize_instance_image_mapping_csv(vms_csv_path: str,
                                          images_csv_path: str,
                                          csv_saving_path: str):
    # Read vms.csv and images.csv
    vms_df = pd.read_csv(vms_csv_path, encoding='utf-8')
    images_df = pd.read_csv(images_csv_path, encoding='utf-8')

    # Assuming GpuTags are stored as string lists
    images_df['GpuTags'] = images_df['GpuTags'].apply(
        lambda x: json.loads(x[1:-1].lower()))

    # Merging on vCenter
    merged_df = pd.merge(vms_df,
                         images_df,
                         left_on='Region',
                         right_on='vCenter')
    results = []

    for _, row in merged_df.iterrows():
        if pd.isna(row['AcceleratorName']) and row['GpuTags'] == []:
            results.append(
                [row['InstanceType'], row['ImageID'], row['vCenter']])
        else:
            if not pd.isna(row['AcceleratorName']):
                vm_gpu_fullname = row['AcceleratorName'].lower()
                vm_gpu_manufacturer_fullname = json.loads(
                    row['GpuInfo'].replace(
                        '\'', '\"'))['Gpus'][0]['Manufacturer'].lower()
                mapping_found_by_gpu_name = False
                default_manufacturer_name = None
                for image_gpu_tag in row['GpuTags']:
                    image_gpu_tag_lower = image_gpu_tag.lower()
                    if image_gpu_tag_lower in vm_gpu_fullname:
                        results.append([
                            row['InstanceType'], row['ImageID'], row['vCenter']
                        ])
                        mapping_found_by_gpu_name = True
                        break
                    elif image_gpu_tag_lower in vm_gpu_manufacturer_fullname:
                        default_manufacturer_name = image_gpu_tag_lower
                if (not mapping_found_by_gpu_name and
                        default_manufacturer_name is not None):
                    results.append(
                        [row['InstanceType'], row['ImageID'], row['vCenter']])
            else:
                for manufacturer in GPU_MANUFACTURERS:
                    if manufacturer in row['GpuTags']:
                        results.append([
                            row['InstanceType'], row['ImageID'], row['vCenter']
                        ])

    result_df = pd.DataFrame(results,
                             columns=['InstanceType', 'ImageID', 'vCenter'])
    result_df.drop_duplicates(inplace=True)
    result_df.to_csv(csv_saving_path,
                     mode='a',
                     header=False,
                     index=False,
                     encoding='utf-8')



import sky

CLOUDS_TO_TEST = [
    'AWS', 'GCP', 'IBM', 'Azure', 'Lambda', 'OCI', 'scp', 'vsphere'
]


def test_list_accelerators():
    result = sky.list_accelerators()
    assert 'V100' in result, result
    assert 'tpu-v3' in result, result
    assert 'Inferentia' not in result, result
    assert 'Trainium' not in result, result
    assert 'A100-80GB' in result, result


def test_list_accelerators_all():
    result = sky.list_accelerators(gpus_only=False)
    assert 'V100' in result, result
    assert 'tpu-v3' in result, result
    assert 'Inferentia' in result, result
    assert 'Trainium' in result, result
    assert 'A100-80GB' in result, result


def test_list_accelerators_name_filter():
    """Test only accelerators with specified name are returned."""
    result = sky.list_accelerators(gpus_only=False, name_filter='V100')
    assert sorted(result.keys()) == ['V100', 'V100-32GB'], result


def test_list_accelerators_region_filter():
    """Test only accelerators in the specified cloud and region are returned."""
    result = sky.list_accelerators(gpus_only=False,
                                   clouds='aws',
                                   region_filter='us-west-1')
    all_regions = []
    for res in result.values():
        for instance in res:
            all_regions.append(instance.region)
    assert all([region == 'us-west-1' for region in all_regions])


def test_list_accelerators_name_quantity_filter():
    """Test only accelerators with specified name and quantity are returned."""
    result = sky.list_accelerators(name_filter='V100', quantity_filter=4)
    all_accelerators = []
    for res in result.values():
        for instance in res:
            all_accelerators.append(
                (instance.accelerator_name, instance.accelerator_count))
    assert all([
        name.__contains__('V100') and quantity == 4
        for name, quantity in all_accelerators
    ])


def test_list_accelerators_positive_quantity_filter():
    """Test only accelerators with specified quantity are returned."""
    result = sky.list_accelerators(quantity_filter=4)
    all_accelerators = []
    for res in result.values():
        for instance in res:
            all_accelerators.append(instance.accelerator_count)
    assert all(quantity == 4 for quantity in all_accelerators)


def test_list_accelerators_name_clouds_filter():
    """Test only accelerators in specified cloud, with specified name are returned."""
    for cloud in CLOUDS_TO_TEST:
        result = sky.list_accelerators(clouds=cloud.lower(), name_filter='V100')
        all_accelerators = []
        for res in result.values():
            for instance in res:
                all_accelerators.append(
                    (instance.accelerator_name, instance.cloud))
        assert all([
            name.__contains__('V100') and acc_cloud == cloud
            for name, acc_cloud in all_accelerators
        ])


def test_list_accelerators_name_quantity_clouds_filter():
    """Test only accelerators in specified cloud, with specified name and quantity are returned."""
    for cloud in CLOUDS_TO_TEST:
        result = sky.list_accelerators(clouds=cloud.lower(),
                                       name_filter='A100',
                                       quantity_filter=4)
        all_accelerators = []
        for res in result.values():
            for instance in res:
                all_accelerators.append(
                    (instance.accelerator_name, instance.cloud,
                     instance.accelerator_count))
        assert all([
            name.__contains__('A100') and acc_cloud == cloud and quantity == 4
            for name, acc_cloud, quantity in all_accelerators
        ])



"""Constants for Paperspace provisioner"""

CPU_INSTANCES_TEMPLATEID = {f'C{i}': 't0nspur5' for i in range(5, 10)}

INSTANCE_TO_TEMPLATEID = {
    'H100x8': 'tvimtol9',
    'H100': 'tiq8mhti',
    'A100-80Gx8': 'tvimtol9',
    'A100-80Gx4': 'tiq8mhti',
    'A100-80Gx2': 'tiq8mhti',
    'A100-80G': 'tiq8mhti',
    'V100-32Gx4': 'twnlo3zj',
    'V100-32Gx2': 'twnlo3zj',
    'V100-32G': 'twnlo3zj',
    'V100': 'twnlo3zj',
    **CPU_INSTANCES_TEMPLATEID
}
NVLINK_INSTANCES = {
    'H100x8',
    'A100-80Gx8',
}

DISK_SIZES = [100, 250, 500, 1000, 2000]



""" RunPod Cloud. """

import json
import typing
from typing import Dict, Iterator, List, Optional, Tuple

from sky import clouds
from sky.clouds import service_catalog
from sky.utils import resources_utils

if typing.TYPE_CHECKING:
    from sky import resources as resources_lib

_CREDENTIAL_FILES = [
    'config.toml',
]


@clouds.CLOUD_REGISTRY.register
class RunPod(clouds.Cloud):
    """ RunPod GPU Cloud

    _REPR | The string representation for the RunPod GPU cloud object.
    """
    _REPR = 'RunPod'
    _CLOUD_UNSUPPORTED_FEATURES = {
        clouds.CloudImplementationFeatures.STOP: 'Stopping not supported.',
        clouds.CloudImplementationFeatures.SPOT_INSTANCE:
            ('Spot is not supported, as runpod API does not implement spot.'),
        clouds.CloudImplementationFeatures.MULTI_NODE:
            ('Multi-node not supported yet, as the interconnection among nodes '
             'are non-trivial on RunPod.'),
        clouds.CloudImplementationFeatures.OPEN_PORTS:
            ('Opening ports is not '
             'supported yet on RunPod.'),
        clouds.CloudImplementationFeatures.IMAGE_ID:
            ('Specifying image ID is not supported on RunPod.'),
        clouds.CloudImplementationFeatures.DOCKER_IMAGE:
            (f'Docker image is currently not supported on {_REPR}.'),
        clouds.CloudImplementationFeatures.CUSTOM_DISK_TIER:
            ('Customizing disk tier is not supported yet on RunPod.'),
        clouds.CloudImplementationFeatures.STORAGE_MOUNTING:
            ('Mounting object stores is not supported on RunPod. To read data '
             'from object stores on RunPod, use `mode: COPY` to copy the data '
             'to local disk.'),
    }
    _MAX_CLUSTER_NAME_LEN_LIMIT = 120
    _regions: List[clouds.Region] = []

    PROVISIONER_VERSION = clouds.ProvisionerVersion.SKYPILOT
    STATUS_VERSION = clouds.StatusVersion.SKYPILOT

    @classmethod
    def _unsupported_features_for_resources(
        cls, resources: 'resources_lib.Resources'
    ) -> Dict[clouds.CloudImplementationFeatures, str]:
        """The features not supported based on the resources provided.

        This method is used by check_features_are_supported() to check if the
        cloud implementation supports all the requested features.

        Returns:
            A dict of {feature: reason} for the features not supported by the
            cloud implementation.
        """
        del resources  # unused
        return cls._CLOUD_UNSUPPORTED_FEATURES

    @classmethod
    def _max_cluster_name_length(cls) -> Optional[int]:
        return cls._MAX_CLUSTER_NAME_LEN_LIMIT

    @classmethod
    def regions_with_offering(cls, instance_type: str,
                              accelerators: Optional[Dict[str, int]],
                              use_spot: bool, region: Optional[str],
                              zone: Optional[str]) -> List[clouds.Region]:
        assert zone is None, 'RunPod does not support zones.'
        del accelerators, zone  # unused
        if use_spot:
            return []
        else:
            regions = service_catalog.get_region_zones_for_instance_type(
                instance_type, use_spot, 'runpod')

        if region is not None:
            regions = [r for r in regions if r.name == region]
        return regions

    @classmethod
    def get_vcpus_mem_from_instance_type(
        cls,
        instance_type: str,
    ) -> Tuple[Optional[float], Optional[float]]:
        return service_catalog.get_vcpus_mem_from_instance_type(instance_type,
                                                                clouds='runpod')

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: str,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[None]:
        del num_nodes  # unused
        regions = cls.regions_with_offering(instance_type,
                                            accelerators,
                                            use_spot,
                                            region=region,
                                            zone=None)
        for r in regions:
            assert r.zones is None, r
            yield r.zones

    def instance_type_to_hourly_cost(self,
                                     instance_type: str,
                                     use_spot: bool,
                                     region: Optional[str] = None,
                                     zone: Optional[str] = None) -> float:
        return service_catalog.get_hourly_cost(instance_type,
                                               use_spot=use_spot,
                                               region=region,
                                               zone=zone,
                                               clouds='runpod')

    def accelerators_to_hourly_cost(self,
                                    accelerators: Dict[str, int],
                                    use_spot: bool,
                                    region: Optional[str] = None,
                                    zone: Optional[str] = None) -> float:
        """Returns the hourly cost of the accelerators, in dollars/hour."""
        del accelerators, use_spot, region, zone  # unused
        return 0.0  # RunPod includes accelerators in the hourly cost.

    def get_egress_cost(self, num_gigabytes: float) -> float:
        return 0.0

    def is_same_cloud(self, other: clouds.Cloud) -> bool:
        # Returns true if the two clouds are the same cloud type.
        return isinstance(other, RunPod)

    @classmethod
    def get_default_instance_type(
            cls,
            cpus: Optional[str] = None,
            memory: Optional[str] = None,
            disk_tier: Optional[resources_utils.DiskTier] = None
    ) -> Optional[str]:
        """Returns the default instance type for RunPod."""
        return service_catalog.get_default_instance_type(cpus=cpus,
                                                         memory=memory,
                                                         disk_tier=disk_tier,
                                                         clouds='runpod')

    @classmethod
    def get_accelerators_from_instance_type(
            cls, instance_type: str) -> Optional[Dict[str, int]]:
        return service_catalog.get_accelerators_from_instance_type(
            instance_type, clouds='runpod')

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        return None

    def make_deploy_resources_variables(
            self,
            resources: 'resources_lib.Resources',
            cluster_name_on_cloud: str,
            region: 'clouds.Region',
            zones: Optional[List['clouds.Zone']],
            dryrun: bool = False) -> Dict[str, Optional[str]]:
        del zones, dryrun  # unused

        r = resources
        acc_dict = self.get_accelerators_from_instance_type(r.instance_type)
        if acc_dict is not None:
            custom_resources = json.dumps(acc_dict, separators=(',', ':'))
        else:
            custom_resources = None

        return {
            'instance_type': resources.instance_type,
            'custom_resources': custom_resources,
            'region': region.name,
        }

    def _get_feasible_launchable_resources(
        self, resources: 'resources_lib.Resources'
    ) -> Tuple[List['resources_lib.Resources'], List[str]]:
        """Returns a list of feasible resources for the given resources."""
        if resources.instance_type is not None:
            assert resources.is_launchable(), resources
            resources = resources.copy(accelerators=None)
            return ([resources], [])

        def _make(instance_list):
            resource_list = []
            for instance_type in instance_list:
                r = resources.copy(
                    cloud=RunPod(),
                    instance_type=instance_type,
                    accelerators=None,
                    cpus=None,
                )
                resource_list.append(r)
            return resource_list

        # Currently, handle a filter on accelerators only.
        accelerators = resources.accelerators
        if accelerators is None:
            # Return a default instance type
            default_instance_type = RunPod.get_default_instance_type(
                cpus=resources.cpus,
                memory=resources.memory,
                disk_tier=resources.disk_tier)
            if default_instance_type is None:
                return ([], [])
            else:
                return (_make([default_instance_type]), [])

        assert len(accelerators) == 1, resources
        acc, acc_count = list(accelerators.items())[0]
        (instance_list, fuzzy_candidate_list
        ) = service_catalog.get_instance_type_for_accelerator(
            acc,
            acc_count,
            use_spot=resources.use_spot,
            cpus=resources.cpus,
            region=resources.region,
            zone=resources.zone,
            clouds='runpod')
        if instance_list is None:
            return ([], fuzzy_candidate_list)
        return (_make(instance_list), fuzzy_candidate_list)

    @classmethod
    def check_credentials(cls) -> Tuple[bool, Optional[str]]:
        """ Verify that the user has valid credentials for RunPod. """
        try:
            import runpod  # pylint: disable=import-outside-toplevel
            valid, error = runpod.check_credentials()

            if not valid:
                return False, (
                    f'{error} \n'  # First line is indented by 4 spaces
                    '    Credentials can be set up by running: \n'
                    f'        $ pip install runpod \n'
                    f'        $ runpod config\n'
                    '    For more information, see https://skypilot.readthedocs.io/en/latest/getting-started/installation.html#runpod'  # pylint: disable=line-too-long
                )

            return True, None

        except ImportError:
            return False, ('Failed to import runpod. '
                           'To install, run: pip install skypilot[runpod]')

    def get_credential_file_mounts(self) -> Dict[str, str]:
        return {
            f'~/.runpod/{filename}': f'~/.runpod/{filename}'
            for filename in _CREDENTIAL_FILES
        }

    @classmethod
    def get_current_user_identity(cls) -> Optional[List[str]]:
        # NOTE: used for very advanced SkyPilot functionality
        # Can implement later if desired
        return None

    def instance_type_exists(self, instance_type: str) -> bool:
        return service_catalog.instance_type_exists(instance_type, 'runpod')

    def validate_region_zone(self, region: Optional[str], zone: Optional[str]):
        return service_catalog.validate_region_zone(region,
                                                    zone,
                                                    clouds='runpod')
